/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*

	Video Playback for WiiU
	(high level interface with Webkit)

*/
#include "config.h"

#if ENABLE(VIDEO) && (PLATFORM(WIIU))

#define MEDIA_DEBUG_NOASSERT
// --------------------------------------------------------------------

#include "MediaPlayerPrivateWiiU.h"
#include "TimeRanges.h"
#include "DomainAccessControlWiiU.h"
#include "GraphicsContext.h"
#include "GraphicsContextPlatformPrivate.h"
#include "Frame.h"
#include "FrameLoaderClientWiiU.h"
#include <PlatformError.h>
#include "WiiUMediaPlugin.h"
#include <plugin/mediaTransportFileHook.h>
#include <plugin/mediaPluginBase.h>
#include <player/mediaPluginPlayer.h>
#include <diagnostic\mediaPerformanceTimer.h>
#include <wkgfx_MP_Core.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>

// --------------------------------------------------------------------

#define	_MEDIAPLAYER_MIN_BUFFERTIME		5.0			// 5 seconds buffer time before playback starts (or all data downloaded if < 5s)
#define	_MEDIAPLAYER_MIN_AUDIODECODED	50			// 50% of the decoded audio buffer must be full to safely start playback
#define	_MEDIAPLAYER_MIN_VIDEODECODED	30			// 30% of the decoded video buffer must be full to safely start playback
#define WORKER_PERIOD		(16.667)

using namespace std;

String	*WebCore::MediaPlayerPrivateWiiU::s_lastQueryTypeP;
String	*WebCore::MediaPlayerPrivateWiiU::s_lastQueryCodecsP;


// special file hook for standard IO
namespace WebCore {
class MEDIAstandardFileHook : public MEDIAtransportFileHook
{
	long long curOffset;
	virtual long long GetFileSize(int fd)
	{
		struct stat st;
		if(fstat(fd, &st) < 0)
			return 0;
		return st.st_size;
	}
	virtual int Open(const char *url, int flag)
	{
		int rwFlags = 0;
		if(flag == READ)
			rwFlags = O_RDONLY;
		else if(flag == WRITE)
			rwFlags = O_WRONLY | O_CREAT;
		else if((flag & (WRITE|READ)) == (WRITE|READ))
			rwFlags = O_RDWR | O_CREAT;
		if(flag & APPEND)
		{
			if(flag & READ)
				rwFlags = O_RDWR | O_CREAT;
			else
				rwFlags = O_WRONLY | O_CREAT;
		}
		if(rwFlags == 0)	// need SOMETHING here
			return -1;
		curOffset = 0;
		return open(url, rwFlags, 0);
	}
	virtual int Read(int fd, long long offset, void *buffer, int count)
	{
		if(offset != curOffset)
		{
			curOffset = offset;
			lseek(fd, curOffset, SEEK_SET);
		}
		int numRead = read(fd, buffer, count);
		curOffset += numRead;
	}
	virtual int Write(int fd, long long offset, void *buffer, int count)
	{
		if(offset != curOffset)
		{
			curOffset = offset;
			lseek(fd, curOffset, SEEK_SET);
		}
		int numWrite = write(fd, buffer, count);
		curOffset += numWrite;
	}
	virtual void Close(int fd)
	{
		if(fd >= 0)
			close(fd);
	}
	virtual int Remove(const char *url)
	{
		return rmdir(url);
	}
};
MEDIAtransportFileHook *sMEDIAFileHook_instanceP;
}

namespace WebCore {
class MediaPlayerPrivateWiiU::BufferMonitor
{
public:
	BufferMonitor()
	{
		Reset();
	}
	~BufferMonitor()
	{
	}
	void Reset()
	{
		mLastBufferStatCheckTime = MEDIAsystemTime::GetInvalidTime();
		mLastVideoBufferedTime = -1.0;
		mLastAudioBufferedTime = -1.0;
	}

	MEDIAsystemTime::Type	mLastBufferStatCheckTime;
	f64						mLastVideoBufferedTime;
	f64						mLastAudioBufferedTime;
};

struct MediaPlayerPrivateWiiU::PeriodicHandleParam
{
	bool	mbRender;
};


MediaPlayerPrivateWiiU::WorkerTimer *MediaPlayerPrivateWiiU::s_workerTimerP;

void MediaPlayerPrivateWiiU::ResetGlobals(void)
{
	MEDIAtransportFileHook *hookP = NULL;
	s_workerTimerP = new WorkerTimer;
	s_lastQueryTypeP = new String;
	s_lastQueryCodecsP = new String;
	// these are two separate globals so that in debugger, it is easier to "watch" them
	hookP = sMEDIAFileHook_instanceP = new MEDIAstandardFileHook;
	MediaBridge::SetTransportFileHook(hookP);
}

void MediaPlayerPrivateWiiU::DestroyGlobals(void)
{
	if (sMEDIAFileHook_instanceP) {
		delete sMEDIAFileHook_instanceP;
		sMEDIAFileHook_instanceP = NULL;
	}
}

MediaPlayerPrivateWiiU::WorkerTimer::WorkerTimer(void)
{
	Reset();
}

void MediaPlayerPrivateWiiU::WorkerTimer::Reset(void)
{
	m_timerP = NULL;
	m_refCount = 0;

	// reset m_players...
	timer.ownerP = this;
	pthread_mutex_init(&m_playersMutex, NULL);
}

bool MediaPlayerPrivateWiiU::WorkerTimer::Setup(void)
{
	if(m_refCount++ == 0)
	{
		if(m_timerP == NULL)
		{
			m_timerP = MediaPlayerPrivateWiiU::CreateTimer(&timer);
			if(m_timerP)
				MediaPlayerPrivateWiiU::StartTimer(m_timerP,(f64)WORKER_PERIOD / 1000.0);
		}
	}
	if(m_timerP)
		return true;
	return false;
}

bool MediaPlayerPrivateWiiU::WorkerTimer::Destroy(void)
{
	if(m_refCount > 0)
	{
		if(--m_refCount == 0)
		{
			if(m_timerP)
			{
				WebCore::MediaPlayerPrivateWiiU::StopTimer(m_timerP);
				WebCore::MediaPlayerPrivateWiiU::DestroyTimer(m_timerP);
				m_timerP = NULL;
			}
			return true;
		}
	}
	return false;
}

void MediaPlayerPrivateWiiU::WorkerTimer::timerCallback(void)
{
	// We may be called by the main event loop long after we shut down as posted timer events can linger in there:
	// This code should savely protect us from doing anything in that case.
	// (Note: Webkit works all in one thread, no need to worry about other race conditions here
	if(m_timerP == NULL || m_refCount == 0)
		return;

	// Ok. Now start with the real work...
	pthread_mutex_lock(&m_playersMutex);
	// Make a local copy of the player array so the handle function is free to add and remove as it sees fit.
	PlayerInfoVector aPlayers(m_players);
	pthread_mutex_unlock(&m_playersMutex);
	for(PlayerInfoVector::iterator it=aPlayers.begin(); it!=aPlayers.end(); ++it)
	{
		WebCore::MediaPlayerPrivateWiiU *pInterface = it->m_pInterface;

		WebCore::MediaPlayerPrivateWiiU::PeriodicHandleParam pX;
		pX.mbRender = true;
		pInterface->handlePeriodic(&pX);
	}
	MediaBridge::UpdateTimer();
}

bool MediaPlayerPrivateWiiU::WorkerTimer::RegisterPlayer(MediaPlayerPrivateWiiU *playerP)
{
	if(Setup() == false)
		return false;
	pthread_mutex_lock(&m_playersMutex);
	m_players.push_back(PlayerInfo(playerP));
	pthread_mutex_unlock(&m_playersMutex);
	return true;
}

bool MediaPlayerPrivateWiiU::WorkerTimer::UnregisterPlayer(MediaPlayerPrivateWiiU *playerP)
{
//	if(m_timerP == NULL)
//		return false;

	pthread_mutex_lock(&m_playersMutex);
	for(PlayerInfoVector::iterator it=m_players.begin(); it!=m_players.end(); ++it)
	{
		if (it->m_pInterface == playerP)
		{
			m_players.erase(it);
			break;
		}
	}
	Destroy();
	pthread_mutex_unlock(&m_playersMutex);
	return true;
}


// --------------------------------------------------------------------

// --------------------------------------------------------------------
/**
	Support function to not have to rebuild the MIME info over and over again
*/
static HashSet<String> mimeTypeCache()
	{
	DEFINE_STATIC_LOCAL(HashSet<String>, cache, ());
	static unsigned long _bTypeListInitialized = false;

	if(!_bTypeListInitialized)
		{
		// Check if we have a listener registered
		if(WiiUMediaInterfacePluginSystem::GetListener())
			{
			WiiUMediaInterfacePluginSystem::GetListener()->BeginAskingForMimeType();
			const char* pMimeType = WiiUMediaInterfacePluginSystem::GetListener()->AskForNextMimeType();
			while(pMimeType	!= 0)
				{
				cache.add(String(pMimeType));	
				pMimeType = WiiUMediaInterfacePluginSystem::GetListener()->AskForNextMimeType();
				}
			WiiUMediaInterfacePluginSystem::GetListener()->EndAskingForMimeType();
			}
		else
			{
			}
		_bTypeListInitialized = true;
		ADD_TO_STATIC_LOCAL(_bTypeListInitialized);
		}
	return cache;
	}

// --------------------------------------------------------------------
/**
	Register the supported MIME types of this media engine
*/
void MediaPlayerPrivateWiiU::getSupportedTypes(HashSet<String>& types)
{
	const HashSet<String>& mimeCache = mimeTypeCache();

	HashSet<String>::const_iterator it = mimeCache.begin();
    HashSet<String>::const_iterator end = mimeCache.end();
    for (; it != end; ++it )
            types.add( *it );
}

bool MediaPlayerPrivateWiiU::isValid(void)
{
	// should do proper checks of allocated data here
	return true;
}

// -------------------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------
/**
	Create an instance of this media engine
*/
PassOwnPtr<MediaPlayerPrivateInterface> MediaPlayerPrivateWiiU::create(MediaPlayer* player)
{
	MediaPlayerPrivateWiiU *p = new MediaPlayerPrivateWiiU(player);
	if(p->isValid() == false)
	{
		delete p;
		return PassOwnPtr<MediaPlayerPrivateInterface>();
	}
	return adoptPtr(p);
}

// --------------------------------------------------------------------------
/**
*/
MediaPlayerPrivateWiiU::MediaPlayerPrivateWiiU(MediaPlayer* player)
    : m_player(player)
	, m_pOplInfo(NULL)
	, m_pPlayer(NULL)
    , m_preLoadMode(MediaPlayer::Auto)
	, m_networkState(MediaPlayer::Empty)
    , m_readyState(MediaPlayer::HaveNothing)
	, m_bMetaDataNew(false)
	, m_bSeeking(false)
	, m_bPaused(true)		// not autoplaying!
	, m_size(IntSize())
	, m_bErrorOccurred(false)
	, m_pCurrentImage(NULL)
	, m_volume(1.0f)
	, m_bMuted(false)
	, m_bLoop(false)
	, m_mediaDuration(-1.0f)
	, m_bEndReached(false)
	, m_seekTime(0.0f)
	, m_bMediaDurationKnown(false)
#if PLATFORM(WIIU)
	, m_drmEnabled(false)
#endif
{
//	m_displayIndex = getDisplayIndex();

	m_pPlayerBufferMonitor = new BufferMonitor;

	pthread_mutex_init(&m_dataMutex, NULL);
}

// --------------------------------------------------------------------------
/**
*/
MediaPlayerPrivateWiiU::~MediaPlayerPrivateWiiU()
{
	s_workerTimerP->UnregisterPlayer(this);
	// now check - player exists?
	if(m_pPlayer)
		m_pPlayer->ReleaseInterface();
	m_pPlayer = NULL;
	delete m_pPlayerBufferMonitor;
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::lockDataMutex()
{
	pthread_mutex_lock(&m_dataMutex);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::unlockDataMutex()
{
#if OS(WINDOWS)
	LeaveCriticalSection(&m_dataMutex);
#else
	pthread_mutex_unlock(&m_dataMutex);
#endif
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::setPreload(MediaPlayer::Preload preLoad)
{
	m_preLoadMode = preLoad;
}

// --------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::stringEndsWith(const std::string & a, const std::string & b)
{
	size_t n = a.size() - b.size();
	if (n < 0)
		return false;
	return strcasecmp(&a.c_str()[n], b.c_str()) == 0;
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::openStream()
{
	// Check if at least the registry could be loaded
	if(MediaBridge::IsRegistryValid() == false)
		 return; // -> Out!

	// Are we already loading or got an error?
	if (m_networkState >= MediaPlayer::Loading)
		return;	// -> Out!

	m_bSeeking = false;
	m_bEndReached = false;

	bool bOk = true;
	
	char tmp[768];
	
	std::string xmlPlayerConfig = "<widget xmlns=\"http://www.w3.org/ns/widgets\">";


	// Close configuration
	xmlPlayerConfig.append("</widget>");
		
	m_pPlayerBufferMonitor->Reset();
	m_pPlayer = MediaBridge::GetPlayer(xmlPlayerConfig.c_str(), this, m_type.c_str());
	if (!m_pPlayer)
	{
		newNetworkState(MediaPlayer::FormatError);
		bOk = false;
	}
	else
	{
		if(m_drmEnabled)
		{
			m_pPlayer->StartDRMSession(m_drmMimeType.utf8().data(), m_drmType.utf8().data(), m_drmXmlConfiguration.utf8().data());
		}
		// try to add to bridge
		s_workerTimerP->RegisterPlayer(this);
		
		if (isURLallowable(m_url))
		{
			// Still all well. Open the URL...
			if (!m_pPlayer->OpenURL(m_url.c_str(), m_type.c_str(), true, m_preLoadMode == MediaPlayer::Auto))
				bOk = false;
		}
		else
			bOk = false;

		if (!bOk)
		{
			newNetworkState(MediaPlayer::FormatError);
			//newNetworkState(MediaPlayer::NetworkError);
		}
	}

	if (bOk)
	{
		// All is well! Switch to buffering state...
		newNetworkState(MediaPlayer::Loading);
		newReadyState(MediaPlayer::HaveNothing);
	}

//TEMP DEBUG
if (!bOk)
	OS::Report(">>>>> DID NOT START LOADING OK (%d, %d)\n", networkState(), readyState());
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::cancelLoad()
{
OS::Report("cancelLoad()\n");
	if (m_networkState < MediaPlayer::Loading || m_networkState == MediaPlayer::Loaded)
		return;	// Not yet loading or all done? -> Out!

	// Delete whatever player we got
	s_workerTimerP->UnregisterPlayer(this);
	if(m_pPlayer)
		m_pPlayer->ReleaseInterface();
	m_pPlayer = NULL;

//CENTRALIZE THIS "RESET" STUFF?
	lockDataMutex();
	m_bMetaDataNew = false;
	m_bPaused = true;
	m_bSeeking = false;
	m_bEndReached = false;
	newNetworkState(MediaPlayer::Empty);
	newReadyState(MediaPlayer::HaveNothing);
	unlockDataMutex();
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::prepareToPlay()
{
//OS::Report("prepareToPlay()\n");
	m_bSeeking = false;
	m_bEndReached = false;
	openStream();
}

#if USE(ACCELERATED_COMPOSITING)
// --------------------------------------------------------------------------
/**
*/
PlatformLayer* MediaPlayerPrivateWiiU::platformLayer() const
{
	return 0;
}
#endif

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::play()
{
//OS::Report("play()\n");
	openStream();
	m_bPaused = false;
	m_bEndReached = false;
}

// --------------------------------------------------------------------------
/**
	Pause playback
*/
void MediaPlayerPrivateWiiU::pause()
{
//OS::Report("pause()\n");
	m_bPaused = true;

	#if OS(WINDOWS)
	m_DEBUG_WasPaused = true;
	#endif
}


/**
    If a HTMLMediaElement is attached to the MediaPlayer, the
    Element will emit the 'ended'-event.
 */
bool MediaPlayerPrivateWiiU::isEnded()
{
	// When we can't determine a duration we check if the player can tell us if he's done.
	if (duration() == numeric_limits<float>::infinity() && 				// duration unknown?
		m_pPlayer && 													// have player?
		m_player->mediaPlayerClient()->mediaPlayerIsLooping() && 		// should playback be looping (if not we don't check with the player!)
		m_pPlayer->HasEnded())											// player says it's done?
		return true;
	if(m_pPlayer->HasEnded())
		return true;
	// Otherwise we compare the playhead against the duration.
	if(m_player->mediaPlayerClient()->mediaPlayerIsLooping() == true)
		return false;
	return currentTime() >= duration();
}

void MediaPlayerPrivateWiiU::timeChanged()
{
	m_player->timeChanged();
}

// --------------------------------------------------------------------------
/**
*/
IntSize MediaPlayerPrivateWiiU::naturalSize() const
{
    if (!hasVideo() || m_readyState < MediaPlayer::HaveMetadata)
        return IntSize();
    return IntSize(m_metaData.m_width, m_metaData.m_height);
}

// --------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::hasVideo() const
{
	return m_pPlayer && (m_pPlayer->HasVideo() == MEDIAinterface::eYes);
}

// --------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::hasAudio() const
{
	return m_pPlayer && (m_pPlayer->HasAudio() == MEDIAinterface::eYes);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::setVisible(bool bVisible)
{
	m_bVisible = bVisible;
}

//---------------------------------------------------------------------------
/**
*/
float MediaPlayerPrivateWiiU::duration() const
{
	if (m_bErrorOccurred || !m_pPlayer)
		return(0.0f);

	//DEBUG -- MEDIA: Videos are always 99 minutes long.
	#if OS(WINDOWS)
	return 998.0f * 60.0f + 59.0f;
	#endif

	f64 dur = m_pPlayer->GetDuration();
	// If we have no known/valid duration, it's as long as it gets...
	if (dur < 0.0)
		return numeric_limits<float>::infinity();
	else
		return float(dur);
}



//---------------------------------------------------------------------------
/**
	Get current time of playback in seconds

	@note	This will return 0.0 if no data is known
*/
float MediaPlayerPrivateWiiU::currentTime() const
{
	if (m_bErrorOccurred || m_pPlayer == NULL)
		return(0.0f);

	if (m_bSeeking)
		return(m_seekTime);

	MEDIApts pts = m_pPlayer->GetCurrentPTS();
	if (!pts.IsValid())
		return(0.0f);

	float cur = pts.GetAsSeconds();
	// Clamp to maximum duration if duration is available.
	float dur = m_pPlayer->GetDuration();
	if (dur > 0.0f && cur > dur)
		cur = dur;
	return(cur);
}



//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::seek(float _time)
{
	if (m_bErrorOccurred || !m_pPlayer)
		return;

	if (m_pPlayer && m_pPlayer->IsSeekSupported())
	{
		f64 time = _time;
		f64 dur = duration();
		f64 pos = currentTime();
		if (time > dur)
			time = dur;
		if (time < 0.0)
			time = 0.0;
		{
			MEDIApluginPlayer::SeekError err = m_pPlayer->SeekTo(time);
			if (err == MEDIApluginPlayer::eSeek_Ok)
			{
				// we need to make sure "current data" continues to be asked for when seeking
				m_seekTime = time;
				m_bSeeking = true;
				m_bEndReached = false;
			}
			else if(err == MEDIApluginPlayer::eSeek_Finished)
			{
				// if the seek finished right away, we actually shouldn't seek anymore
				m_seekTime = time;
				m_bSeeking = true;
				m_bEndReached = false;
				newReadyState(MediaPlayer::HaveEnoughData);
			}
		}
	}
}

//---------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::seeking() const
{
    return m_bSeeking;
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::setRate(float rate)
{
	if(m_pPlayer)
		m_pPlayer->SetPlaybackRate(rate);
// Not supported
}

//---------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::paused() const
{
    return m_bPaused;
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::setVolume(float v)
{
	m_volume = v;
	if(m_pPlayer)
	{
		if (!m_bMuted)
			m_pPlayer->SetVolume(m_volume);
		else
			m_pPlayer->SetVolume(0);
	}
}

//---------------------------------------------------------------------------
PassRefPtr<OutputProtectionLevel> MediaPlayerPrivateWiiU::outputProtectionLevel() const
{
	// for this case, let's ask the player
	if(m_pPlayer)
	{
		const MEDIAdrmOplInfo *infoP = m_pPlayer->GetOutputProtectionLevel();
		if(infoP)
			return OutputProtectionLevel::create(infoP->mCompressedDigitalVideo, infoP->mUncompressedDigitalVideo, infoP->mAnalogVideo, infoP->mCompressedDigitalAudio, infoP->mUncompressedDigitalAudio);
	}
	return PassRefPtr<OutputProtectionLevel>();//OutputProtectionLevel::create();
}

//---------------------------------------------------------------------------
bool MediaPlayerPrivateWiiU::startDrmSession(const String& mimeType, const String& drmType, const String& xmlConfiguration)
{
	m_drmMimeType = mimeType;
	m_drmType = drmType;
	m_drmXmlConfiguration = xmlConfiguration;
	m_drmEnabled = true;
	return true;
}

void MediaPlayerPrivateWiiU::cancelDrmSession(void)
{
	m_drmEnabled = false;
}

//---------------------------------------------------------------------------
void MediaPlayerPrivateWiiU::setDeviceVolume(float volume, int id)
{
	if(m_pPlayer)
		m_pPlayer->SetDeviceVolume(volume, (MEDIAaudioDeviceControl::OutputDevice)id);
}

int MediaPlayerPrivateWiiU::getNumChannels(void)
{
	if(m_pPlayer)
		return m_pPlayer->GetNumChannels();
	return 0;
}

void MediaPlayerPrivateWiiU::setChannelVolume(int channel, float volume)
{
	if(m_pPlayer)
		m_pPlayer->SetChannelVolume(channel, volume);
}

void MediaPlayerPrivateWiiU::setChannelPan(int channel, float pan)
{
	if(m_pPlayer)
		m_pPlayer->SetChannelPan(channel, pan);
}

bool MediaPlayerPrivateWiiU::isFinished(void)
{
	if(m_pPlayer)
		return m_pPlayer->HasEnded();
	return false;
}

bool MediaPlayerPrivateWiiU::canRestart(void)
{
	if(m_pPlayer)
		return true;
	return false;
}

void MediaPlayerPrivateWiiU::restart(void)
{
	bool backupVolume = false;
	int i;
	f32 volume;
	f32 deviceVolume[MEDIAaudioDeviceControl::kNumDevices];
	if(m_pPlayer)
	{
		backupVolume = true;
		volume = m_pPlayer->SetVolume(0);
		for(i = 0; i < MEDIAaudioDeviceControl::kNumDevices; i++)
		{
			deviceVolume[i] = m_pPlayer->SetDeviceVolume(0, (MEDIAaudioDeviceControl::OutputDevice)i);
		}
	}
	cancelLoad();
	openStream();
	if(m_pPlayer && backupVolume)
	{
		m_pPlayer->SetVolume(volume);
		for(i = 0; i < MEDIAaudioDeviceControl::kNumDevices; i++)
		{
			m_pPlayer->SetDeviceVolume(deviceVolume[i], (MEDIAaudioDeviceControl::OutputDevice)i);
		}
	}
}


//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::setMuted(bool bMute)
{
	if (bMute != m_bMuted)
	{
		if (m_pPlayer)
		{
			if (bMute)
				m_pPlayer->SetVolume(0.0f);
			else
				m_pPlayer->SetVolume(m_volume);
		}
		m_bMuted = bMute;
	}
}

void MediaPlayerPrivateWiiU::setLoop(bool bLoop)
{
	if(bLoop != m_bLoop)
	{
		if(m_pPlayer)
		{
			m_pPlayer->SetLoop(bLoop);
		}
	}
}


//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::setErrorState()
{
	m_bErrorOccurred = true;
}

//---------------------------------------------------------------------------
/**
*/
MediaPlayer::NetworkState MediaPlayerPrivateWiiU::networkState() const
{
    return m_networkState;
}

//---------------------------------------------------------------------------
/**
*/
MediaPlayer::ReadyState MediaPlayerPrivateWiiU::readyState() const
{
    return m_readyState;
}

//-----------------------------------------------------------------------------
/**
*/
float MediaPlayerPrivateWiiU::maxTimeSeekable() const
{
	if (m_bErrorOccurred || !m_pPlayer)
		return 0.0f;

	if (!m_pPlayer->IsSeekSupported())
		return 0.0f;

	float dur = duration();
	return isfinite(dur) ? dur : 0.0f;
}

//---------------------------------------------------------------------------
/**
*/
PassRefPtr<TimeRanges> MediaPlayerPrivateWiiU::buffered() const
{
    RefPtr<TimeRanges> timeRanges = TimeRanges::create();

	// We are not strictly speaking buffering anything. We do live content streaming.
	// As such we do not have anything buffered. For best compatibility we say we have everything buffered up.
	f32 dur = duration();
	if (!m_bErrorOccurred && m_pPlayer && isfinite(dur))
		timeRanges->add(0, dur);
    return timeRanges.release();
}

//---------------------------------------------------------------------------
/**
*/
unsigned MediaPlayerPrivateWiiU::bytesLoaded() const
{
	if (m_pPlayer)
		return m_pPlayer->GetBufferedSize();
	return 0;
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::setSize(const IntSize& size)
{
    m_size = size;
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::loadingFailed(MediaPlayer::NetworkState error)
{
	m_bErrorOccurred = true;
	newNetworkState(error);
	newReadyState(MediaPlayer::HaveNothing);
}


//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::useImage(const MEDIAdecodedVideoFrame *pCurrentRenderImage)
{
	m_pCurrentImage = pCurrentRenderImage;
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::paint(GraphicsContext* context, const IntRect& rect)
{
	// Please see comment at method: MediaPlayerPrivateWiiU::naturalSize()
	if (context->paintingDisabled() || !m_bVisible || !hasVideo() || isEnded())
        return;

	// Receive the current image from the player
	m_pCurrentImage = NULL;
	MEDIAvideoRenderer::RenderParams	params;
	if (m_pPlayer)
		m_pPlayer->VideoRender(params);

	if (m_pCurrentImage)
	{
		MEDIAdecodedVideoFrame::Dimension	dim;
		MEDIAdecodedVideoFrame::Crop		crop;
		MEDIAdecodedVideoFrame::ImageData	tex;

		m_pCurrentImage->GetDimension(dim);
		m_pCurrentImage->GetCrop(crop);
		m_pCurrentImage->GetImageData(tex);
		ASSERT(tex.mpPlane2 == 0 && m_pCurrentImage->GetImageFormat() == MEDIAdecodedVideoFrame::eFmt_YUV);
		// We got image data! Setup texture objects and ask context to render the data...
		FloatRect src(crop.mLeft, crop.mTop, dim.mW - crop.mRight, dim.mH - crop.mBottom);
		context->platformContext()->drawDataYUV(FloatRect(rect), src, (void*)tex.mpPlane0, (void*)tex.mpPlane1, dim.mW, dim.mH);
		m_pCurrentImage = NULL;
	}
	else
	{
		// No data. A simple black rectangle must do...
		Color black(0,0,0,255);
		context->fillRect(FloatRect(rect), black, ColorSpaceDeviceRGB);
	}
}

//-----------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::hasSingleSecurityOrigin() const
{
    // We tell to disallow resources that come from different origins so we are sure all media is single origin
    return true;
}

//-----------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::newReadyState(MediaPlayer::ReadyState newState)
{
	bool bRet;
	if (m_readyState == newState)
		bRet = false;
	else
	{
		m_readyState = newState;
		m_player->readyStateChanged();
		bRet = true;
	}
	return bRet;
}

//-----------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::newNetworkState(MediaPlayer::NetworkState newState)
{
	bool bRet;
	if (m_networkState == newState)
		bRet = false;
	else
	{
		m_networkState = newState;
		m_player->networkStateChanged();
		bRet = true;
	}
	return bRet;
}

//-----------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::updateMetaData()
{
	lockDataMutex();
	if (m_bMetaDataNew)
	{
		m_bMetaDataNew = false;

		if (m_readyState < MediaPlayer::HaveMetadata)
			newReadyState(MediaPlayer::HaveMetadata);
	}
	unlockDataMutex();
}

//-----------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::setMetaData(const MediaBridge::MetaData & metaData)
{
	// This is called from outside the webkit thread, we hence need to push the info
	// to be later on routed into the webkit world during the timer call...
	lockDataMutex();
	m_bMetaDataNew = true;
	m_metaData = metaData;
	unlockDataMutex();
}

//-----------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::repaint()
{
	m_player->repaint();
}

//-----------------------------------------------------------------------------
/**
	Determine if playback could be (reasonably) safely started right now

	@note	Do only use this to determine if you should go from "no playback" to
			"playback". Generally "playback" should continue until data truly,
			simply runs out! This check is way more conservative!
*/
int MediaPlayerPrivateWiiU::haveEnoughData(MEDIApluginPlayer *pPlayer)
{
	if(pPlayer == NULL)
		return false;
	if(pPlayer->IsEnoughDataReady())	// enough data received, allow playback
		return -1;
	return false;
}



void MediaPlayerPrivateWiiU::handlePeriodic(PeriodicHandleParam *pParam)
{
	bool bMaybeRestartForLooping = false;
	bool bShouldUpdateAfterSeek = false;

	if (!m_pPlayer)
		return;
	// if the OPL is set, trigger an event
	if(m_pPlayer->GetOutputProtectionLevel() != m_pOplInfo)
	{
		m_pOplInfo = m_pPlayer->GetOutputProtectionLevel();
		if(m_pOplInfo)
			m_player->outputProtectionLevelChanged();
	}

	MEDIAerrorBase *pError = m_pPlayer->GetError();
	if (pError)
	{
		// Generally note down that there was an error
		setErrorState();


		switch (pError->GetArea())
			{
			case	MEDIAerrorBase::AREA_TRANSPORT:
				newNetworkState(WebCore::MediaPlayer::NetworkError);
				break;
			case	MEDIAerrorBase::AREA_DECODE:
			case	MEDIAerrorBase::AREA_DEMUX:
				newNetworkState(WebCore::MediaPlayer::DecodeError);
				break;
			default:
				{
				// ? nothing to post ?
				}
			}

		m_pPlayer->ReleaseError(pError);
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	// Give the player interface a chance to push out any updates reg. meta data
	updateMetaData();

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
/*
Removed since this has an undesired side effect: telling that everything is loaded prevents
an instance of our class to be destroyed!

This is caused by (a bug in?) WebCore::HTMLMediaElement::userCancelledLoad() where they leave
the function when the file has been completely loaded:
	if (m_networkState == NETWORK_EMPTY || m_completelyLoaded)

Since the function is left early the player is not destroyed a couple of lines later:
	m_player.clear();


	if (m_pPlayer->IsAllDataReceived())
	{
		// Ok. We got all there is to get...
		newNetworkState(WebCore::MediaPlayer::Loaded);
	}
*/
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	int enough;

	// Do we think we have enough data?
	if (readyState() < WebCore::MediaPlayer::HaveEnoughData)
	{
		// No. Do we have enough now?
		if (enough = haveEnoughData(m_pPlayer))
		{
			// Yes! Tell the world we got it...
			newReadyState(WebCore::MediaPlayer::HaveEnoughData);

			// ...and then, if we should, start playback...
			if (!paused())
			{
				if (!m_pPlayer->IsRendering())
					m_pPlayer->StartRendering();
			}
			if (m_bSeeking)
				bShouldUpdateAfterSeek = true;
			m_bSeeking = false;
		}
	}
	else
	{
		// Theoretically we should still have enough data. First a bit state checking...

		// Ok. So what should our playback state be?
		if (paused() || m_bEndReached)
		{
			// We should be paused. If we still play, stop things now...
			if (m_pPlayer->IsRendering())
			{
				if(m_bEndReached == false && paused())
				{
					m_pPlayer->PauseRendering();
				}
				else
					m_pPlayer->StopRendering();
			}
		}
		else
		{
			// PTS-wise, are we ended?
			if (isEnded())
			{
				// Yes!
				if (!m_bEndReached)
				{
					m_bEndReached = true;
//					pause();
					timeChanged();
					m_pPlayer->StopRendering();
					bMaybeRestartForLooping = true;
				}
			}
			// We should be playing. Are we?
			else if (!m_pPlayer->IsRendering())
			{
				// We are not! Did we reach the end?
				if (m_pPlayer->HasEnded())
				{
					m_bEndReached = true;
					// Well, pause the stream. Nothing more to come...
//					pause();
				}
				// We are not rendering, but we are not at the end, do we no longer have enough data?
				else
				{
					enough = haveEnoughData(m_pPlayer);
					if (enough)
					{
						// Yes. Even though the ready state was in "have enough data", send it again?  Only if seeking
						if(m_bSeeking)
						{
							newReadyState(WebCore::MediaPlayer::HaveEnoughData);
							bShouldUpdateAfterSeek = true;
							m_bSeeking = false;
						}
						// Well, there seems to be no reason not to be playing! We got data after all...
						m_pPlayer->StartRendering();
					}
					else
					{
						// we actually no longer have data... let WebKit know we don't
						newReadyState(WebCore::MediaPlayer::HaveMetadata);	// we don't have current data, we are back to META DATA
					}
				}
			}
			else
			{
				// we did not end, we ARE rendering... maybe we were seeking?
				enough = haveEnoughData(m_pPlayer);	// find out how much data we have?
				if(enough)
				{
					if (m_bSeeking)	// we are seeking and we have enough data, send an update to let them know it's ok
					{
						newReadyState(WebCore::MediaPlayer::HaveEnoughData);	// let webkit know the seek is done
						bShouldUpdateAfterSeek = true;
						m_bSeeking = false;
					}
				}
				else	// we no longer have data, switch states
				{
					if(m_pPlayer->IsRendering())
						m_pPlayer->PauseRendering();
					newReadyState(WebCore::MediaPlayer::HaveMetadata);
				}
			}
		}
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	// Keep track of the display we are on so we get audio from the right set of speakers
/*	int dispIdx = getDisplayIndex();
	if (dispIdx >= 0)
	{
		// 0 is the TV, 1 is the GamePad...
		m_pPlayer->SetAudioDevice((dispIdx == 0) ? MEDIAaudioDeviceControl::kDevice_TV : MEDIAaudioDeviceControl::kDevice_Gamepad,
								(dispIdx == 0) ? 0 : dispIdx - 1);
	}*/

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	// Trigger rendering as needed...
	if (pParam->mbRender && hasVideo() && !isEnded())
	{
		// We display the movie right on the webkit context. Trigger a paint event from webkit for us...
		repaint();
	}

	if (bShouldUpdateAfterSeek)
		timeChanged();

	// Restart for looping?
	if (bMaybeRestartForLooping && m_player->mediaPlayerClient()->mediaPlayerIsLooping())
	{
		if(m_pPlayer->IsSeekSupported())
		{
			m_pPlayer->SeekTo(0);
			m_pPlayer->StartRendering();
		}
		else
		{
			restart();
		}
	}

}



#if USE(ACCELERATED_COMPOSITING)

//-----------------------------------------------------------------------------
/**
*/
// whether accelerated rendering is supported by the media engine for the current media.
bool MediaPlayerPrivateWiiU::supportsAcceleratedRendering() const
{
    return false;
}

//-----------------------------------------------------------------------------
/**
*/
// called when the rendering system flips the into or out of accelerated rendering mode.
void MediaPlayerPrivateWiiU::acceleratedRenderingStateChanged()
{
}

//-----------------------------------------------------------------------------
/**
*/
// Callback for when hardware-accelerated animation started.
void MediaPlayerPrivateWiiU::notifyAnimationStarted(const GraphicsLayer*, double time)
{
}

//-----------------------------------------------------------------------------
/**
*/
// Notification that a layer property changed that requires a subsequent call to syncCompositingState()
// to appear on the screen.
void MediaPlayerPrivateWiiU::notifySyncRequired(const GraphicsLayer*)
{
}

//-----------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::paintContents(const GraphicsLayer*, GraphicsContext&, GraphicsLayerPaintingPhase, const IntRect& inClip)
{
}

//-----------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::didCommitChangesForLayer(const GraphicsLayer*) const
{
}

//-----------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::showDebugBorders() const
{
    return false;
}

//-----------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::showRepaintCounter() const
{
    return false;
}

#endif

// --------------------------------------------------------------------
/**
	Check if this media engine is available
*/
bool MediaPlayerPrivateWiiU::isAvailable()
{
	// In this context we assume always available player
	return true;
}

// --------------------------------------------------------------------
/**
	Check if we support a given type
*/
MediaPlayer::SupportsType MediaPlayerPrivateWiiU::supportsType(const String& type, const String& codecs)
{
	*s_lastQueryTypeP = type;
	*s_lastQueryCodecsP = codecs;

	if (type.isNull() && type.isEmpty())
		return MediaPlayer::IsNotSupported;

    // spec says we should not return "probably" if the codecs string is empty
    if (mimeTypeCache().contains(type))
        return codecs.isEmpty() ? MediaPlayer::MayBeSupported : MediaPlayer::IsSupported;

	return MediaPlayer::IsNotSupported;
}

// --------------------------------------------------------------------
/**
	Register the media engine with Webkit
*/
void MediaPlayerPrivateWiiU::registerMediaEngine(MediaEngineRegistrar registrar)
{
	if (isAvailable())
        registrar(create, getSupportedTypes, supportsType, 0, 0, 0);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::load(const String& url)
{
	m_url = encodeWithURLEscapeSequences(url).latin1().data();

	m_type = s_lastQueryTypeP->utf8().data();
	m_codecs = s_lastQueryCodecsP->utf8().data();

OS::Report("Load %s MIME=%s CODECS=%s\n", m_url.c_str(), m_type.c_str(), m_codecs.c_str());

	if (m_preLoadMode == MediaPlayer::Auto)
		openStream();
}

// --------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::isURLallowable(const std::string & url)
	{
			int error = WebCore::DomainAccessControl::DOMAIN_ACCESS_OK;
	if (!WebCore::DomainAccessControl::IsDomainAccessible(url.c_str(), &error))
			{
				switch (error)
				{
				case WebCore::DomainAccessControl::DOMAIN_ACCESS_NO_NETWORK_ACCESS:
					OS::Report("This request cannot be honored. You must enable network access.\n");
					LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_ASSET_LOAD_FAIL, L"", 0);
					break;

				case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_ENCRYPTED:
					OS::Report("This request must be over HTTPS. Please use the HTTPS protocol.\n\tRequested URL: %s.\n", url.c_str());
					LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_CURLE_UNSUPPORTED_PROTOCOL, L"", 0);
					break;

				case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_WHITELISTED:
					OS::Report("The url '%s' is not whitelisted. Add its domain to config.xml to whitelist.\n", url.c_str());
					LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_ADDRESS_BLOCKED_BY_WHITELIST, L"", 0);
					break;

				case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_ACQUIRED_AUTH_TOKEN:
					OS::Report("Before accessing a network resource, an independent service token must first be acquired.\n" \
							"Please call nwf.act.NintendoAccountManager.acquireNewIndependentServiceToken().\n");
					break;

				default:
					OS::Report("This domain is unreachable. Error %d\n", error);
					break;
				}
				return false;
			}
	return true;
}

// --------------------------------------------------------------------------
/**
	Return index of display the player is on

	@return	Index of display (TV=0, GAMEPAD=1,2...); returns -1 if unknown
*/

int MediaPlayerPrivateWiiU::getDisplayIndex() const
{
	FrameLoader *pFrameLoader = m_player->mediaPlayerClient()->mediaPlayerOwningDocument()->frame()->loader();
	if (!pFrameLoader)
		return -1;

	WebView *pView = static_cast<FrameLoaderClientWiiU*>(pFrameLoader->client())->webView();

	if (pView->GetDisplayType() == WebView::GAMEPAD_DISPLAY_TYPE)
		return 1 + pView->GetGamePadIndex();
	return 0;
}

// --------------------------------------------------------------------------
/**
*/
Timer<MediaTimer>* MediaPlayerPrivateWiiU::CreateTimer(MediaTimer *pTimer)
{
	return new Timer<MediaTimer>(pTimer, &MediaTimer::_timerCallback);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::DestroyTimer(Timer<MediaTimer>* pTimer)
{
	delete pTimer;
}

// --------------------------------------------------------------------------
/**
*/
void  MediaPlayerPrivateWiiU::StartTimer(Timer<MediaTimer>* pTimer, double period)
{
	pTimer->startRepeating(period);
}

// --------------------------------------------------------------------------
/**
*/
void  MediaPlayerPrivateWiiU::StopTimer(Timer<MediaTimer>* pTimer)
{
	pTimer->stop();
}

// --------------------------------------------------------------------------
/**
 */
void MediaTimer::_timerCallback(WebCore::Timer<MediaTimer>*)
{
	timerCallback();
}


// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

WiiUMediaInterfacePluginSystem::Listener* WiiUMediaInterfacePluginSystem::mpListener = NULL;
		
void WiiUMediaInterfacePluginSystem::SetListener(WiiUMediaInterfacePluginSystem::Listener* pListener)
	{ 
	mpListener = pListener;
	}
	
WiiUMediaInterfacePluginSystem::Listener* WiiUMediaInterfacePluginSystem::GetListener(void)
	{
	return mpListener;
	}

}; // namespace WebCore


#endif

