/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef JSSYSTEMEVENTDISPATCHER_H
#define JSSYSTEMEVENTDISPATCHER_H

#include <wtf/Platform.h> // PLATFORM(...)

#include "JSEventSystem.h"
#include <PlatformInput.h>

struct WiiUKeyData {
	signed channel;
	unsigned key_state;
};

struct WiiUScalarData {
	signed channel;
	float s;
};

struct WiiUVector2Data {
	signed channel;
	float x;
	float y;
};

struct WiiUVector3Data {
	signed channel;
	float x;
	float y;
	float z;
};

struct WiiUVector4Data {
	signed channel;
	float x;
	float y;
	float z;
	float w;
};

struct WiiUAccelData {
	signed channel;
	float dir_x;
	float dir_y;
	float dir_z;
	float mag;
	float change;
};

struct WiiUMoveData {
	signed channel;
	float x;
	float y;
	float z;
	float screenX;
	float screenY;
};

struct WiiUGyroData {
	signed channel;
	float angle_x;
	float angle_y;
	float angle_z;
	float vel_x;
	float vel_y;
	float vel_z;
};

struct WiiUErrorData {
	std::wstring description;
	std::string originFunction;
	unsigned code;	
	unsigned type;
	unsigned originCode;
};

enum WiiUSystemDevice {
	REMOTE = 0,
	REMOTE_CL,
	REMOTE_FS,
	PRO,
	GAMEPAD,
	CAM,
	IO,
	AOC,
	ACT,
	SYSTEM,
	ERROR_WINDOW,
	DISPLAY,
	BALANCEBOARD,
	MII,
	MIIVERSE,
	DATA_STORE,
	FRIEND_PRESENCE,
	NETWORK,
	GAME_SERVER,
	RANKING,
    SOFTWAREKEYBOARD,
	DEVICE_COUNT
};

enum WiiUSystemChannel {
	CHANNEL_0 = 0,
	CHANNEL_1,
	CHANNEL_2,
	CHANNEL_3,
	CHANNEL_COUNT
};

enum WiiUSystemControl {
	DEFAULT = 0,
	BUTTON,
	MOVEMENT,
	DIRECTION,
	GYRO,
	LSTICK,
	RSTICK,
	ACCELEROMETER,
	DPD,
	TOUCH,
	WEIGHT,
	WEIGHT_CORRECTED,
    MIC,
	POST,
	CONTROL_COUNT
};

class JSExtensionEvent: public JSEvent
{
public:
	JSExtensionEvent(WiiUEventType_t event_id, std::string event_type, void *data = NULL, std::string event_string = "", 
		WiiUSystemDevice deviceID = REMOTE, WiiUSystemControl controlID = DEFAULT, WiiUSystemChannel indexID = CHANNEL_0);

	WiiUSystemChannel GetIndexID();
	WiiUSystemDevice GetDeviceID();
	WiiUSystemControl GetControlID();
private:
	WiiUSystemChannel indexID;
	WiiUSystemDevice deviceID;
	WiiUSystemControl controlID;
};

// Depending on the application different input data is recorded and used.
// We group different types of input together in their own functions like buttons, gyro/accel, stick, touch screen, etc.
// Doing this allows us to only detect input that is necessary for the application.
// If nothing requires gyro data we dont bother detecting and dispatching its data.
typedef void (*DispatchFunction)(signed channel);
class DispatchFunctor {
public:
	DispatchFunctor(DispatchFunction function, signed channel, std::string event_type) : function(function) , channel(channel), event_type(event_type) {}
	~DispatchFunctor(void) {}

	void operator ()() {
		this->function(this->channel);
	}

	DispatchFunction function;
	signed channel;
	std::string event_type;
};

/**
@brief This is a special dispatcher that must be listened to in order to recieve system events without polling for them from js.
*/
class JSSystemEventDispatcher : public JSEventDispatcher {
public:
	struct KeyDispatcherPair {
		KeyDispatcherPair(void) : key(""), dispatcher(0) {}
		KeyDispatcherPair(std::string key, DispatchFunction dispatcher) : key(key), dispatcher(dispatcher) {}
		~KeyDispatcherPair();

		std::string key;
		DispatchFunction dispatcher;
	};

	struct ListenerEventPair {
		ListenerEventPair(void) : js_listener(0) {}
		ListenerEventPair(JSEventListener *js_listener, std::string event_type) : js_listener(js_listener), event_type(event_type) {}
		JSEventListener *js_listener;
		std::string event_type;
	};

	struct ListenerEventFunctorTuple {
		ListenerEventFunctorTuple(
			JSEventListener *js_listener, 
			std::string event_type, 
			DispatchFunctor functor) : 
			js_listener(js_listener),
			event_type(event_type),
			functor(functor) {}

		JSEventListener *js_listener;
		std::string event_type;
		DispatchFunctor functor;
	};

private:
	JSSystemEventDispatcher(void);
	virtual ~JSSystemEventDispatcher(void);

public:
	static JSSystemEventDispatcher *GetInstance(void);
	static void DestroyInstance(void);
	
	bool AddDefaultEventListener(JSEventListener *js_listener, std::string event_type);

	// Now with less string concat!
	virtual bool AddEventListener(
		JSEventListener *js_listener, 
		WiiUSystemDevice device, 
		WiiUSystemChannel channel,
		WiiUSystemControl control,
		std::string event_type
	);

	virtual bool AddEventListener(
		JSEventListener *js_listener, 
		WiiUSystemDevice device, 
		std::string event_type
	);

	virtual void RemoveEventListener(JSEventListener *js_listener, std::string event_type);

	virtual void RemoveAllListeners(JSEventListener *js_listener);

	unsigned int HasEventListener(std::string event_type);
	bool HasEventListener(JSEventListener *js_listener, std::string event_type);

	void DispatchInputEvents(void);
	void DispatchSystemEvent(std::string event_name, std::string event_string, void *data, WiiUEventType_t event_type = WIIU_EMPTY_EVENT);
	void DispatchExtensionEvent(JSExtensionEvent* jsExtensionEvent, std::string event_string);
	void DispatchExtensionEventThreadSafe(JSExtensionEvent* jsExtensionEvent, std::string event_string);
	void DispatchExtensionEvent(std::string event_name, std::string event_string, void *data, WiiUSystemChannel indexID = CHANNEL_0, WiiUSystemDevice deviceID = REMOTE, WiiUSystemControl controlID = DEFAULT);
	void DispatchErrorEvent(JSEvent* js_event);
	void DispatchDisplayEvent(std::string event_string, WiiUSystemChannel channel);
	void DispatchCrashEvent(void);
	
	void JSSystemEventDispatcher::QueueEventForDispatch(
		std::string event_name, 
		WiiUSystemDevice device, 
		WiiUSystemChannel channel, 
		WiiUSystemControl control, 
		JSEvent *event
	);

	std::string GetEventString(WiiUSystemDevice device, WiiUSystemChannel channel, WiiUSystemControl control, std::string event_type);
	std::string GetEventString(WiiUSystemDevice device, std::string event_type);

	void SetCrash(bool crash);
	bool GetCrash(void);
	void SetCrashCallback(void (*callback)(void));

private:
	std::vector<DispatchFunctor> dispatch_functors;
	std::vector<DispatchFunctor *> dispatch_functor_pointers;
	//std::vector<std::vector<std::vector<std::map<std::string, KeyDispatcherPair> > > > event_strings;
	std::map<unsigned int, KeyDispatcherPair> event_strings[DEVICE_COUNT][CHANNEL_COUNT][CONTROL_COUNT];
	std::vector<ListenerEventPair> remove_list;
	std::vector<ListenerEventPair> listener_add_list;
	std::vector<ListenerEventFunctorTuple> add_list;
	std::vector<JSEventListener *> clear_list;
	bool crashing;
	bool crash_event_sent;
	void (*crash_callback)(void);
	pthread_mutex_t	addListMutex;
	pthread_mutex_t eventListenerMapMutex;
};

#endif//JSSYSTEMEVENTDISPATCHER_H
